package net.sf.sido.factory.json;

import net.sf.sido.DOJsonFieldNotArrayException;
import net.sf.sido.DataObject;
import net.sf.sido.factory.support.AbstractDataFactoryReader;
import net.sf.sido.factory.support.SiDOSchema;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.codehaus.jackson.JsonNode;
import org.codehaus.jackson.JsonParser;
import org.codehaus.jackson.map.ObjectMapper;

import java.io.IOException;
import java.net.URL;
import java.util.Iterator;

import static net.sf.sido.factory.json.JSONUtils.getBoolean;
import static net.sf.sido.factory.json.JSONUtils.getString;

public class JSONDataFactoryReader extends AbstractDataFactoryReader {

    @Override
    public DataObject readSchema(SiDOSchema sidoSchema, URL url) throws IOException {
        // Checks
        Validate.notNull(url, "The schema URL is required");
        Validate.notNull(sidoSchema, "The schema factory is required");
        // Factory
        ObjectMapper mapper = new ObjectMapper();
        mapper.configure(JsonParser.Feature.ALLOW_UNQUOTED_FIELD_NAMES, true);
		mapper.configure(JsonParser.Feature.ALLOW_COMMENTS, true);
        // Reads
        JsonNode jSchema = mapper.readValue(url, JsonNode.class);

        // Gets the schema name
        String name = getString(jSchema, "name", true, null);
        DataObject schema = sidoSchema.createSchema(name);

        // Gets the types
        JsonNode jTypes = jSchema.get("types");
        if (jTypes != null) {
            // Reads all types
            readTypes(sidoSchema, schema, jTypes);
        }

        // Reads all extensions
        JsonNode jExtensions = jSchema.get("extensions");
        if (jExtensions != null) {
            // Reads all extensions
            readExtensions(sidoSchema, schema, jExtensions);
        }

        // OK
        return schema;
    }

    protected void readExtensionProperty(SiDOSchema sido, DataObject extension, JsonNode jProperty) {
        readProperty(sido, extension, jProperty);
    }

    protected void readExtension(SiDOSchema sido, DataObject schema, JsonNode jExtension) {
        // Gets the type to extend
        String parentName = getString(jExtension, "parent", true, null);
        DataObject extension = sido.createExtension(schema, parentName);

        // 1 - simple declaration
        if (jExtension.get("name") != null) {
            readExtensionProperty(sido, extension, jExtension);
        }

        // 2 - list of extensions
        JsonNode jProperties = jExtension.get("properties");
        for (JsonNode jProperty : jProperties) {
            readExtensionProperty(sido, extension, jProperty);
        }
    }

    protected void readExtensions(SiDOSchema sido, DataObject schema, JsonNode jExtensions) {
        // Checks this is an array
        if (!jExtensions.isArray()) {
            throw new DOJsonFieldNotArrayException("extensions");
        }
        // For each extension
        for (JsonNode jExtension : jExtensions) {
            readExtension(sido, schema, jExtension);
        }
    }

    protected void readType(SiDOSchema sido, DataObject schema, JsonNode jType) {
        // Gets the type name
        String name = getString(jType, "name", true, null);
        // Creates the type
        DataObject type = sido.createType(schema, name);
        // Extensibility
        if (JSONUtils.getBoolean(jType, "extensible", false, false)) {
            type.setBoolean(SiDOSchema.EXTENSIBLE, true);
        }
        // Abstract type
        if (JSONUtils.getBoolean(jType, "abstract", false, false)) {
            type.setBoolean(SiDOSchema.ABSTRACT, true);
        }
        // Parent type
        String parent = JSONUtils.getString(jType, "parent", false, null);
        if (StringUtils.isNotBlank(parent)) {
            type.setString(SiDOSchema.PARENT, parent);
        }
        // Reads the properties
        JsonNode jProperties = jType.get("properties");
        if (jProperties != null) {
            if (!jProperties.isArray()) {
                throw new DOJsonFieldNotArrayException("properties");
            }
            for (JsonNode jProperty : jProperties) {
                readProperty(sido, type, jProperty);
            }
        }
    }

    protected void readProperty(SiDOSchema sido, DataObject type, JsonNode jProperty) {
        DataObject property = sido.createProperty(type);
        property.setString(SiDOSchema.NAME, getString(jProperty, "name", true, null));
        property.setBoolean(SiDOSchema.NULLABLE, getBoolean(jProperty, "nullable", false, false));
        property.setBoolean(SiDOSchema.REVERSE, getBoolean(jProperty, "reverse", false, false));
		property.setString(SiDOSchema.ARRAY_INDEX, getString(jProperty, "index", false, null));
        // Type
        String jType = getString(jProperty, "type", false, "string");
        boolean anonymous;
        boolean array;
        String typeref;
        if ("#".equals(jType)) {
            // Anonymous type
            anonymous = true;
            array = false;
            typeref = null;
        } else if ("*#".equals(jType)) {
            // Array of generics
            anonymous = true;
            array = true;
            typeref = null;
        } else if (jType.startsWith("*#")) {
            // Array of DOs
            anonymous = false;
            array = true;
            typeref = jType.substring(2);
        } else if (jType.startsWith("*")) {
            // Array of simple types
            anonymous = false;
            array = true;
            typeref = jType.substring(1);
        } else if (jType.startsWith("#")) {
            // DO
            anonymous = false;
            array = false;
            typeref = jType.substring(1);
        } else {
            // Simple
            anonymous = false;
            array = false;
            typeref = jType;
        }
        property.setBoolean(SiDOSchema.ANONYMOUS, anonymous);
        property.setBoolean(SiDOSchema.ARRAY, array);
        property.setString(SiDOSchema.TYPEREF, typeref);
        // Constraints
        readConstraints(sido, property, jProperty);
    }

    protected void readConstraints(SiDOSchema sido, DataObject property, JsonNode jProperty) {
        JsonNode jConstraint = jProperty.get("constraint");
        if (jConstraint != null && !jConstraint.isNull()) {
            // List of constraint
            if (jConstraint.isArray()) {
                for (JsonNode jsonNode : jConstraint) {
                    readConstraint(sido, property, jsonNode);
                }
            }
            // Simple constraint
            else {
                readConstraint(sido, property, jConstraint);
            }
        }
    }

    protected void readConstraint(SiDOSchema sido, DataObject property, JsonNode jConstraint) {
        String id = getString(jConstraint, "type", true, null);
        DataObject constraint = sido.createConstraint(property, id);
        // All parameters
        Iterator<String> fieldNames = jConstraint.getFieldNames();
        while (fieldNames.hasNext()) {
            String fieldName = fieldNames.next();
            if (!"type".equals(fieldName)) {
                sido.setConstraintParam(constraint, fieldName, jConstraint.get(fieldName).getValueAsText());
            }
        }
    }

    protected void readTypes(SiDOSchema sido, DataObject schema, JsonNode jTypes) throws IOException {
        // Checks this is an array
        if (!jTypes.isArray()) {
            throw new DOJsonFieldNotArrayException("types");
        }
        // Count of types
        int count = jTypes.size();
        for (int i = 0; i < count; i++) {
            // Type definition
            JsonNode jType = jTypes.get(i);
            // Reads the type
            readType(sido, schema, jType);
        }
    }
}
